package ses;

import java.util.ArrayList;

import repast.simphony.context.Context;
import repast.simphony.essentials.RepastEssentials;
import animals.Animal;

/**
 * Interface for the java wraper of the scala rangeland model
 * @author karma
 *
 */
public class EcoSpace 
{
	private BiomassSimulator iBiomass;
	private double gfd = 0.7;
	private double haRangeland = 2500;
	private ArrayList<Double> monthlyLwgs = new ArrayList<Double>();
	private double yearlyLwg;
	private ArrayList<Double> monthlyBA = new ArrayList<Double>();
	
	
	/**
	 * Initializes the rangeland with continous grazing
	 * @param theContext
	 */
	public EcoSpace(Context theContext)
	{
		iBiomass= new BiomassModel(3, 1);
		theContext.add(iBiomass);
		iBiomass.initRun();
	}

	/**
	 * initializes the rangeland without grazing in monthly time steps
	 * @param currentMonth
	 */
	public void init(int currentMonth) 
	{
		iBiomass.stepMonth(currentMonth, 0, 0);
	}
	
	/**
	 * Steps the rangeland with offtake and biomass growth.
	 * @param currentMonth
	 * @param allAnimals
	 */
	public void stepMonthlyBiomass(int currentMonth,ArrayList<Animal> allAnimals )
	{
		double offtake = getOfftake(allAnimals);		
		iBiomass.stepMonth(currentMonth, offtake, gfd);	
		assert(iBiomass.getTSC()>=0&&iBiomass.getGSC()>=0&&iBiomass.getGGrowth()>=0);	
		double gfd = LiveWeightGain.getGreenInDiet(iBiomass.getTSC(), iBiomass.getGSC());
		assert(gfd<=1&&gfd>=0);
		this.gfd=gfd;
		
		monthlyBA.add(iBiomass.getBasalArea());
	}
	
	/**
	 * 
	 * @return total standing crop kg/ha
	 */
	public double getTSC()
	{
		return iBiomass.getTSC();
	}
	
	/**
	 * 
	 * @return green standing crop  kg/ha
	 */
	public double getGSC()
	{
		return iBiomass.getGSC();
	}
	
	/**
	 * 
	 * @return green growth kg/ha
	 */
	public double getGGrowth()
	{
		return iBiomass.getGGrowth();
	}
	
	/**
	 * 
	 * @return the real share of green fodder in the diat. Diverges from wished depending on gsc availability
	 */
	public double getActualGFD() 
	{
		return iBiomass.getActualGFD();
	}
	
	/**
	 * 
	 * @return actual forage consumed. Diverges from wished offtake depending on forage avaialbility
	 */
	public double getActualOfftake()
	{
		return iBiomass.getActualOfftake();
	}
	
	/**
	 * 
	 * @return basal area
	 */
	public double getBasalArea()
	{
		return iBiomass.getBasalArea();
	}
	
	/**
	 * Calculates the offtake of animals depending on body-weight and stocking density per month
	 * @param allAnimals
	 * @return
	 */
	public double getOfftake(ArrayList<Animal> allAnimals)
	{
		double bodyMass = getTotalBodyMass(allAnimals);
		double bodyMassAvr = bodyMass/allAnimals.size();		
		double intkCoeff = 0.02;
		double days = 31;
		double stockingDensity = (bodyMass/250)/haRangeland;
		double intkMult = 1;
		
		double result = bodyMassAvr*intkCoeff*days*stockingDensity*intkMult;
		
		if(allAnimals.isEmpty())
			return 0;
		else
			return result;
	}
	
	/**
	 * Gets the total body mass of all animals on the rangeland
	 * @param allAnimals
	 * @return
	 */
	private double getTotalBodyMass(ArrayList<Animal> allAnimals)
	{
		double sum=0;
		for(Animal anAnimal:allAnimals)
		{
			sum=sum+anAnimal.getBodyWeight();
		}
		
		double bodyMass;
		if (allAnimals.size() == 0)
			bodyMass = 0;
		else
			bodyMass = sum;
		
		return bodyMass;
	}
	
	/**
	 * gets the liveweight gain of the past month
	 * @return
	 */
	public double getLiveweightgain()
	{
		double tsc = getTSC();
		double gsc = getGSC();
		double ggrth = getGGrowth();
		double actGfd = getActualGFD();
		
		return  LiveWeightGain.calculateLWG(tsc, gsc, ggrth, actGfd);
	}
	
	/**
	 * Stores monthly live-weight gains or losses in a list
	 * @param lwg
	 */
	public void storeMonhtlyLwg(double lwg)
	{
		monthlyLwgs.add(lwg);
	}
	
	/**
	 * calculates the sum of monthly lwg the array of monthly lwgs
	 * @return
	 */
	public double calcYearlyLwg()
	{
		double sum = 0;
		for(Double aLwg: monthlyLwgs)
		{
			sum+=aLwg;
		}		
		return sum;
	}
	
	/**
	 * Gets the yearly live-weight gain or loss
	 * @return
	 */
	public double getYearlyLwg()
	{
		return yearlyLwg;
	}
	
	/**
	 * helper method reseting yearly live-weight gain to zero for a new year
	 */
	public void resetYearlyLwg()
	{
		monthlyLwgs.clear();
	}
	
	/**
	 * memory method for storing last years lwg
	 * @param lwg
	 */
	public void setLwgLastYear(double lwg)
	{
		yearlyLwg=lwg;
	}
	
	/**
	 * Finishes the run and of the rangeland model
	 */
	public void finishRunBiom()
	{
		iBiomass.finishRun();
		iBiomass.finish();
	}
//---------------------------Repast only output methods-----------------------------------------------//	
	public double getStockingRate(ArrayList<Animal> allAnimals)
	{
		double bodyMass = getTotalBodyMass(allAnimals);
		double stockingDensity = (bodyMass/400)/haRangeland;
		
		return stockingDensity;
	}
	
	/**
	 * only returning the value stored for the last year. No calculations here
	 * @return
	 */
	public double getLastYearLwg()
	{
		return yearlyLwg;
	}
	
	public double calcMeanBA()
	{
		double sum=0;
		for(Double aBa : monthlyBA)
		{
			sum+=aBa;
		}
		return sum/monthlyBA.size();
	}

	public double calcStDevBA()
	{
		double var=0;
		double mean = calcMeanBA();
        double temp = 0;
        for(double aBa :monthlyBA)
            temp += (mean-aBa)*(mean-aBa);
            
        var = temp/monthlyBA.size();
        
        return Math.sqrt(var);
	}
	
	public double getOfftake()
	{
		return iBiomass.getActualOfftake();
	}
	
	public double getOfftakeByTSC()
	{
		double offtakebyTSC = iBiomass.getActualOfftake()/iBiomass.getTSC();
		if(offtakebyTSC>1)
			offtakebyTSC=1;
		return offtakebyTSC;
	}
	
	public double getOfftakeByGSC()
	{
		return iBiomass.getActualOfftake()/iBiomass.getGSC();
	}
	
}
