package ses;

import java.util.ArrayList;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.parameter.Parameters;
import agents.ProducerHH;

/**
 * The CollectiveAction is the social medium of normative interaction of participants. Participants engage in collective action
 * by normative sanctioning and norm obedience or defection. 
 * @author karma
 *
 */
public class SocialSpace 
{
	
	public SocialSpace()
	{
		Parameters params = RunEnvironment.getInstance().getParameters();
		maxSize =(Double) params.getValue("maxHerdSize");
	}
	
	/**
	 * The inverted salience: Salience is the importance, urgency or conciseness of a situation. 
	 * In Sediba this is determined by inequality aversion. High inequality among participants increases the salience
	 * of the situation. We use the inverted salience (1-gini) as a threshold for conformity to a norm and normative sanctioning.
	 */
	private double invertVividness = 0.5;
	
	/**
	 * Severity describes the degree of current violations of the norm. For Sediba it is the sum of cattle  above the
	 * maximum herd size per household. 
	 */
	private double severity=0;
	
	/**
	 * Group trust describes the trust of participants in the fact that others will cooperate 
	 * and that norm violations will be punished by normative sanctioning
	 */
	private double groupRepuation=0.5;
	
	/**
	 * The maximimum herd size per HH. Greater herd sizes per HH would lead to a destruction of the resource
	 */
	private double maxSize=25;
	
	/**
	 * The percentage of defecting agents in relation to all participants. This includes also principle defectors 
	 * (not only defecting weavers)
	 */
	private double levelOfDefection=1-groupRepuation;
	

	/**
	 * The current stocking rate
	 */
	private double totalSize=500;
	
	
	/**
	 * nr of permanent defectors in the system
	 */
	private double percDefectors;


	
	/**
	 * Main step method of the CollectiveAction. It updates the severity of the non-conformity to a maximum herd size, updates 
	 * the group level of trust that defectors will be punished and that participants will stick to the norm, updates the 
	 * principle conditional cooperativeness of individual agents, updates the salience of the situation which is determined
	 * by the inequality between herd sizes and steps the agents to check whether they will actually engage in the collective
	 * action.
	 * @param giniHerds the inequality (Gini coefficient) between participant's (including defectors) herd sizes
	 */
	public void stepCollectiveAction(double giniHerds, ArrayList<ProducerHH> allProducers)
	{
		updateLevelOfDefection(allProducers);
		updateGroupReputationReciprocity();
		updateGroupReputationCooperation(allProducers);	
		updateCooperativeness(allProducers);
		updatedNormSanction(giniHerds);
		checkPartCooperation(allProducers);				
	}
	
	/**
	 * Updates the current level of defection. An agent defects if his conditional cooperativeness is below the 
	 * vividness of the situation
	 */
	private void updateLevelOfDefection(ArrayList<ProducerHH> allProducers)
	{
		double observedDefectors=0;
		double permDefectors=0;
		for(ProducerHH aProducer :allProducers)
		{
			if(aProducer.isDefector())
			{
				observedDefectors++;
				permDefectors++;
			}
			else if(aProducer.isCooperator()==false)
			{
				observedDefectors++;
			}
		}

		percDefectors = permDefectors/allProducers.size();
		levelOfDefection = observedDefectors/allProducers.size();
	}
	
	/**
	 * Updates the trust (or reputation of the group )that others will reciprocitate hostile acts (normative sanctioning). 
	 * Trust that others will be reciprocators is determined by past trust and the percentage of reciprocators relative to all
	 * participants in the action situation. This lagged effect mimiks group reputation building.
	 */
	private void updateGroupReputationReciprocity()
	{		
		groupRepuation= (1-levelOfDefection);		
	}
	
	/**
	 * Updates the trust (or repuation of the group) that others will cooperate in case that the norm is in effect 
	 * in their case. That is, expected cooperativenss is influenced by the degree of existing norm violations.
	 * Norm violations are measured as the % of animals which are above the max herd size
	 * per HH in relation to the total stocking rate.
	 */
	private void updateGroupReputationCooperation(ArrayList<ProducerHH> allProducers)
	{
		double illegalAnimals=0;
		double totalSize=0;
		for(ProducerHH aParticipant:allProducers)
		{
			totalSize+=aParticipant.getHerdSize();
			if(aParticipant.getHerdSize()>maxSize)
			{
				illegalAnimals+=aParticipant.getHerdSize()-maxSize;
			}
		}
		  severity=illegalAnimals/totalSize;		
	}
	
	/**
	 * Updates the attributes of agents which are used in the calculation for their conditional cooperativeness. 
	 */
	private void updateCooperativeness(ArrayList<ProducerHH> allProducers)
	{
		for(ProducerHH aParticipant:allProducers)
		{
			aParticipant.updateCooperation(groupRepuation, severity);
		}
	}
	
	private void updatedNormSanction(double giniHerds)
	{
			invertVividness= (invertVividness+(1-(giniHerds)))/2;	
	}
	
	private void checkPartCooperation(ArrayList<ProducerHH> allProducers)
	{
		for(ProducerHH aParticipant:allProducers)
		{
			if(aParticipant.isDefector())
			{
				aParticipant.setCooperator(false);
			}
			else
			{
				double condCooperativeness = aParticipant.getConditionalCooperation();
				if(condCooperativeness>invertVividness)
					aParticipant.setCooperator(true);
				else
					aParticipant.setCooperator(false);
			}			
		}
	}
	
	public void setMaxSize(double maxSize) 
	{
		this.maxSize = maxSize;
	}

	public double getGroupTrust()
	{
		return 1-levelOfDefection;
	}
	
	public double getNormSanct()
	{
		return 1-invertVividness;
	}
	
	public double getMaxSize()
	{
		return maxSize;
	}
	
	public double getSeverity()
	{
		return severity;
	}
	
	public void setTotalSize (double totalSize)
	{
		this.totalSize=totalSize;
	}

	public double getPercDefectors() 
	{
		return percDefectors;
	}	
}
